var proxy = {
    debug: true,
    modules: {
        crypto: require('crypto'),
        dnsd: require('dnsd'),
        exec: require('child_process').exec,
        fs: require('fs'),
        http: require('http'),
        https: require('https'),
        tls: require('tls'),
        url: require('url'),
        uuid: require('uuid'),
    },

    requests: [], // List of in-progress requests
    dnsServerEndpoint: false,
    httpServerEndpoint: false,
    httpsKeys: {}, // Stores HTTPS keys for various domains
    httpsOptions: {
        SNICallback: function (domain, callback) {
//            console.log('Checking SNI', domain);
            if (proxy.httpsKeys[domain]) {
                return callback(null, proxy.httpsKeys[domain]);
            }

            // Generate keys on the fly
            proxy.generateHttpsKeys(domain).then(function() {
                return callback(null, proxy.httpsKeys[domain]);
            });
        },
        key: false,
        cert: false,
    },
    serverEndpointHttps: false,

    init: function() {
        // Network exchange folder
        if (proxy.modules.fs.existsSync('network') === false) {
            proxy.modules.fs.mkdirSync('network');
        }

        proxy.modules.fs.truncateSync('/root/scratch/vm/ca/index.txt');

        proxy.initDnsServerEndpoint();
        proxy.initHttpServerEndpoint();

        try {
            // Generate default key on the fly
            proxy.generateHttpsKeys('charityengine.com').then(function() {
                proxy.httpsOptions.key = proxy.modules.fs.readFileSync('/root/scratch/vm/ca/certs/charityengine.com.key.pem', 'utf8');
                proxy.httpsOptions.cert = proxy.modules.fs.readFileSync('/root/scratch/vm/ca/certs/charityengine.com.pem', 'utf8');
                proxy.initHttpsServerEndpoint();
            });
        } catch (e) {
            console.log('Unable to init HTTPS', e);
        }
    },
    initDnsServerEndpoint: function() {
        proxy.dnsServerEndpoint = proxy.modules.dnsd.createServer(function(req, res) {
            //console.log(req);
            //proxy.newRequest(req, res, 'dns');
            console.log('DNS query ('+req.connection.type+')', req.question);
            res.end('127.0.0.1');
        }).listen(53, '0.0.0.0');
        console.log('DNS server running at 0.0.0.0:53');

        proxy.dnsServerEndpoint2 = proxy.modules.dnsd.createServer(function(req, res) {
            console.log('dns 2');
            console.log(req);
            proxy.newRequest(req, res, 'dns');
        }).listen(3645, '0.0.0.0');
        console.log('listening dns');
    },
    initHttpServerEndpoint: function() {
        proxy.httpServerEndpoint = proxy.modules.http.createServer(function (req, res) {
            let id = proxy.newRequest(req, res, 'http');
            setTimeout(proxy.monitorRequest, 1000, id);
        })
            .listen(3646, '0.0.0.0', 1023);
        console.log('listening http');
    },
    initHttpsServerEndpoint: function() {
        proxy.serverEndpointHttps = proxy.modules.https.createServer(proxy.httpsOptions, function (req, res) {
            let id = proxy.newRequest(req, res, 'https');
            setTimeout(proxy.monitorRequest, 1000, id);
        })
            .listen(3647, '0.0.0.0', 1023);
        console.log('listening https', 3647);
    },
    monitorRequest: function(uuid) {
        //console.log('checking request', uuid);
        let file = uuid + '.res';
        proxy.modules.fs.access('network/' + file, proxy.modules.fs.constants.R_OK, function (err) {
            if (!err) {
//                console.log('response received', uuid);
                // Decrypt the request
                proxy.modules.fs.readFile('network/' + file, function(err, data) {
                    let dataString = data.toString();
                    let splitLocation = dataString.indexOf(':');
                    let iv = Buffer.from(dataString.substr(0, splitLocation), 'hex');
                    let encrypted = data.slice(splitLocation + 1);

                    let decipher = proxy.modules.crypto.createDecipheriv('aes-256-ctr', new Buffer('60ac35dd1573b0be115171f79d6630ba'), iv);
                    let decrypted = decipher.update(encrypted);
                    decrypted += decipher.final();
                    let response = JSON.parse(decrypted);
                    proxy.requests[uuid].resolve(response);

                    // Cleanup network folder
                    proxy.modules.fs.unlink('network/' + uuid + '.res', function() {}); // Empty callback, ignore errors
                });
            }
            else {
                setTimeout(proxy.monitorRequest, 1000, uuid);
            }
        });
    },
    finalizeRequest: function (id, req, res, request) {
        console.log('[SERVER] New request ' + id + ' for ' + req.url);
        console.log(request.data);

        proxy.requests[id] = request;
        let encryptedRequest = proxy.encryptRequestData(request.data);
        proxy.modules.fs.writeFile('network/' + id + '.req', encryptedRequest, function () {
        });

        var promise = new Promise(function(resolve, reject) {
            proxy.requests[id].resolve = resolve;
            proxy.requests[id].reject = reject;
        });
        promise.then(
            function(response) {
                response = JSON.parse(response);
                console.log('request is done', id);
                res.writeHead(response.statusCode, response.headers);
                res.end(Buffer.from(response.content, 'base64'));
            },
            function(err) {
                console.log('request failed', err);
                res.end();
            }
        );
    },
    newRequest: function(req, res, type) {
        console.log('new connection', type);
        let parsedUrl = proxy.modules.url.parse(req.url);

        let request = {
            req: req,
            res: res,
            data: { // Must be compatible with node.js http.request options object
                metadata: {
                    type: type,
                },
                object: {
                    headers: req.headers,
                    hostname: (parsedUrl.hostname ? parsedUrl.hostname : req.headers.host),
                    method: req.method,
                },
            },
            resolve: null,
            reject: null,
        };

        if (typeof req.path !== 'undefined') {
            request.data.object.path = req.path;
        } else if (typeof req.url !== 'undefined') {
            if (req.url.substr(0, 1) !== '/') {
                console.log('using url', req.url);
                process.exit();
            }
            request.data.object.path = req.url;
        }

        switch (type) {
        case 'https':
            request.data.object.hostname = req.headers.host;
            break;
        }

        let id = proxy.modules.uuid.v4();

        switch (req.method) {
            case 'POST':
                var postData = '';
                req.on('data', function (data) {
                    postData += data;
                });
                // Limit POST data
                if (postData.length > 1000000) {
                    req.connection.destroy();
                }
                req.on('end', function () {
                    request.data.object.postData = postData;
                    if (proxy.debug) {
                        console.log('Post data', postData);
                    }
                    proxy.finalizeRequest(id, req, res, request);
                });
                break;
            default:
                proxy.finalizeRequest(id, req, res, request);
                break;
        }

        return id;
    },
    encryptRequestData: function(data) {
        let iv = proxy.modules.crypto.randomBytes(16);
        let cipher = proxy.modules.crypto.createCipheriv('aes-256-ctr', new Buffer('60ac35dd1573b0be115171f79d6630ba'), iv);
        let encrypted = cipher.update(JSON.stringify(data));
        let finalBuffer = Buffer.concat([encrypted, cipher.final()]);
        return iv.toString('hex') + ':' + finalBuffer.toString('base64');
    },
    generateHttpsKeys: function(domain) {
        return new Promise(function(resolve, reject) {
            // @todo Needs error checking
            proxy.modules.exec('openssl genrsa -out /root/scratch/vm/ca/certs/' + domain + '.key.pem 2048', function callback(error, stdout, stderr) {
                //console.log(error, stdout, stderr);
                proxy.modules.exec('openssl req -config /root/scratch/vm/ca/openssl.cnf -key /root/scratch/vm/ca/certs/' + domain + '.key.pem -new -sha256 -out /root/scratch/vm/ca/certs/' + domain + '.csr.pem -subj \'/CN=' + domain + '/O=Proxy/C=US/ST=Massachusetts\'', function callback(error, stdout, stderr) {
                    //console.log(error, stdout, stderr);
                    proxy.modules.exec('openssl ca -batch -config /root/scratch/vm/ca/openssl.cnf -extensions server_cert -days 365 -notext -md sha256 -in /root/scratch/vm/ca/certs/' + domain + '.csr.pem -out /root/scratch/vm/ca/certs/' + domain + '.pem', function callback(error, stdout, stderr) {
                        //console.log(error, stdout, stderr);
                        proxy.httpsKeys[domain] = proxy.modules.tls.createSecureContext({
                            key: proxy.modules.fs.readFileSync('/root/scratch/vm/ca/certs/' + domain + '.key.pem', 'utf8'),
                            cert: proxy.modules.fs.readFileSync('/root/scratch/vm/ca/certs/' + domain + '.pem', 'utf8'),
                        });
                        resolve(proxy.httpsKeys[domain]);
                    });
                });
            });
        });
    }
}

module.exports = proxy;

proxy.init();
setTimeout(function() {}, 100000);
